import os
import sys
import distutils.spawn
import subprocess
import threading

from .utils import default_wait_func


__all__ = ['find_file', 'IterProcess', 'pip_bin']


def find_file(*filenames, default=None):
    """Find a file using the python paths.

    Args:
        *filenames (tuple/str): File names to look for.
        default (str)[None]: If not found return this result.

    Returns:
        filename (str)[None]: Filename that was found or default argument.
    """
    # Check normal path
    for fname in filenames:
        try:
            # Should work on windows and linux to find executable/binaries like pip.exe
            exe = distutils.spawn.find_executable(fname)
            if exe is not None:
                return exe
        except (ValueError, TypeError, Exception):
            pass
        try:
            if os.path.exists(fname):
                return fname
        except (ValueError, TypeError, Exception):
            pass

    for path in sys.path:
        for fname in filenames:
            try:
                file = os.path.join(path, fname)
                if os.path.exists(file):
                    return file
            except (ValueError, TypeError, Exception):
                pass

    return default


class IterProcess(subprocess.Popen):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._buffer = []
        self._toggle = 'out'
        self.stopped_out = False
        self.stopped_err = False
        try:
            self.iter_out = iter(self.stdout.readline)  # Iterator that yields data or ''
        except (AttributeError, Exception):
            self.iter_out = None
        try:
            self.iter_err = iter(self.stderr.readline)  # Iterator that yields data or ''
        except (AttributeError, Exception):
            self.iter_err = None

        self.th_out = threading.Thread(target=self.thread_read_buffer, args=(self.stdout,), daemon=True)
        self.th_err = threading.Thread(target=self.thread_read_buffer, args=(self.stderr,), daemon=True)
        self.th_out.start()
        self.th_err.start()

    def is_running(self):
        """Return if the process is still running."""
        return self.poll() is None

    def get_sys_file(self, file):
        """Return sys.stdout or sys.stderr based on the given file object buffer."""
        if file == self.stdout:
            return sys.stdout
        else:
            return sys.stderr

    def thread_read_buffer(self, file):
        """Read the buffer in a separate thread.

        Buffer reading is blocking, so this must be done in a separate thread.
        """
        while self.is_running():
            try:
                line = file.readline()
                if isinstance(line, bytes):
                    line = line.decode('utf-8', 'replace')
                self._buffer.append((line, file))
            except:
                pass

        # Try for longer?
        for _ in range(100):
            try:
                line = file.readline()
                if isinstance(line, bytes):
                    line = line.decode('utf-8', 'replace')
                self._buffer.append((line, file))
            except (ValueError, TypeError, Exception):
                break  # On error stop

    def next(self):
        try:
            line, file = self._buffer.pop(0)
        except (IndexError, ValueError, TypeError, Exception) as err:
            line, file = '', None

        if not line and not self.is_running():
            raise StopIteration
        return line, file

    def close(self):
        try:
            self.stdout.close()
        except (AttributeError, Exception):
            pass
        try:
            self.stderr.close()
        except (AttributeError, Exception):
            pass
        try:
            self.th_out.join()
        except (AttributeError, Exception):
            pass
        try:
            self.th_err.join()
        except (AttributeError, Exception):
            pass

    def __next__(self):
        return self.next()

    def __iter__(self):
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return exc_tb is None


def pip_bin(*args, wait_func=None, **kwargs):
    """Run pip using the binary/executable.

    https://pip.pypa.io/en/stable/reference/

    Example:

        >>> pip_bin('install', '-e', '../mylib')

    Args:
        *args (tuple/str): Arguments that are normally passed into pip.
        wait_func (callable/function): Call this function while waiting for pip to finish.
        **kwargs (dict/object)[None]: Catch extra arguments

    Returns:
        exitcode (int): Process exit code
    """
    pip_main_func = find_file('pip.exe', 'pip3', 'pip', default=None)
    if pip_main_func is None:
        raise EnvironmentError('Cannot find the pip binary!')

    if wait_func is None:
        wait_func = default_wait_func

    try:
        args = [pip_main_func] + list(args)
        if 'stdout' not in kwargs:
            kwargs['stdout'] = subprocess.PIPE
        if 'stderr' not in kwargs:
            kwargs['stderr'] = subprocess.PIPE

        with IterProcess(args, **kwargs) as proc:
            for line, file in proc:
                wait_func()
                if line:
                    print(line, end='', file=proc.get_sys_file(file))
            # print('finishing')

        return getattr(proc, 'returncode', 1)
    except (ValueError, TypeError, OSError, Exception) as err:
        return 1



import sys
import inspect
import multiprocessing as mp
import light_process as lp

from pylibimport.run_pip.utils import default_wait_func


try:
    from pip._internal.cli.main import main as PIP_MAIN_FUNC
except (ImportError, AttributeError, Exception):
    try:
        from pip._internal import main as PIP_MAIN_FUNC
    except (ImportError, AttributeError, Exception):
        try:
            from pip import main as PIP_MAIN_FUNC
        except (ImportError, AttributeError, Exception):
            PIP_MAIN_FUNC = None  # Not available for some reason.


__all__ = ['PIP_MAIN_FUNC', 'is_pip_main_available', 'pip_main', 'is_pip_proc_available', 'pip_proc', 'pip_proc_flag']


def is_pip_main_available():
    """Return if the main pip function is available. Call get_pip_main before calling this function."""
    return PIP_MAIN_FUNC is not None


is_pip_proc_available = is_pip_main_available


def pip_main(*args, **kwargs):
    """Run a pip command using the pip main function.

    https://pip.pypa.io/en/stable/reference/

    Note:
        pip_main may spawn another process. This may make executables not work.

    Example:

        >>> pip_main('install', '-e', '../mylib')

    Args:
        *args (tuple/str): Arguments that are normally passed into pip.
        **kwargs (dict/object)[None]: Catch extra arguments

    Raises:
        EnvironmentError: If pip could not be imported

    Returns:
        exitcode (int): Process exit code
    """
    if not is_pip_main_available():
        raise EnvironmentError('The main pip function is not available for this environment! Try pip_bin.')
    return PIP_MAIN_FUNC(list(args))


def pip_proc(*args, wait_func=None, **kwargs):
    """Run a pip command using multiprocessing.

    https://pip.pypa.io/en/stable/reference/

    Note:
        pip_main may spawn another process. This may make executables not work.

    Example:

        >>> pip_proc('install', '-e', '../mylib')

    Args:
        *args (tuple/str): Arguments that are normally passed into pip.
        wait_func (callable/function): Call this function while waiting for pip to finish.
        **kwargs (dict/object)[None]: Catch extra arguments

    Raises:
        EnvironmentError: If pip could not be imported

    Returns:
        exitcode (int): Process exit code
    """
    if not is_pip_proc_available():
        raise EnvironmentError('The main pip function is not available for this environment! Try pip_bin.')

    if wait_func is None:
        wait_func = default_wait_func

    try:
        # Calling pip_main is bad practice (could do undesirable things). Run it in another process ...
        proc = lp.LightProcess(target=PIP_MAIN_FUNC, args=(list(args), ))  # , name='pip_main')
        proc.start()
        while proc.exitcode is None:
            wait_func()
        proc.join(1)

        return getattr(proc, 'exitcode', 1)
    except (ValueError, TypeError, OSError, Exception) as err:
        return 1


def _pip_proc_flag(*args, finished_flag=None, success_flag=None, **kwargs):
    """Run pip and flag when it is finished."""
    ret = 1
    try:
        ret = PIP_MAIN_FUNC(list(args))
    finally:
        if ret == 0 and success_flag is not None:
            success_flag.set()
        if finished_flag is not None:
            finished_flag.set()
    return ret


def pip_proc_flag(*args, wait_func=None, **kwargs):
    """Old multiprocessing does not successfully use exitcode to check if the process is finished."""
    if wait_func is None:
        wait_func = default_wait_func

    try:
        kwargs['finished_flag'] = finished_flag = mp.Event()
        kwargs['success_flag'] = success_flag = mp.Event()
        proc = mp.Process(target=_pip_proc_flag, args=args, kwargs=kwargs)
        proc.start()
        while not finished_flag.is_set():
            wait_func()
        proc.join(1)
        try:
            proc.terminate()  # Make sure it closed
        except:
            pass

        exitcode = getattr(proc, 'exitcode', 1)
        if success_flag.is_set():
            exitcode = 0
        return exitcode
    except(ValueError, TypeError, Exception):
        return 1



import os
import sys
import copy
import shutil
import tarfile
import contextlib
import importlib
from collections import OrderedDict

from pylibimport.utils import get_name_version
from pylibimport.run_pip import default_wait_func, pip_main


__all__ = ['InstallError', 'original_system', 'import_module', 'install_lib',
           'register_install_type', 'remove_install_type', 'get_install_func',
           'is_python_package', 'py_install', 'is_zip', 'zip_install', 'whl_install']


class InstallError(Exception):
    pass


@contextlib.contextmanager
def original_system(new_path=None, reset_modules=True):
    """Context manager to reset sys.path and sys.modules to the previous state before the context operation.

    Args:
        new_path (str)[None]: Temporarily add a path to sys.path before the operation.
        reset_modules (bool)[True]: If True reset sys.modules back to the original sys.modules.
    """
    modules = sys.modules.copy()
    paths = copy.copy(sys.path)
    path_cache = sys.path_importer_cache.copy()

    # Temporarily add the new path
    if new_path and new_path not in sys.path:
        sys.path.insert(0, new_path)

    yield

    if reset_modules:
        # sys.modules = modules  # For some reason this causes conflicts with relative imports?
        sys.modules.clear()
        sys.modules.update(modules)
    sys.path = paths
    sys.path_importer_cache = path_cache


def import_module(path, import_chain=None, reset_modules=True):
    """Import the given module name from the given import path.

    Args:
        path (str): Directory which contains the module to import.
        import_chain (str): Chain to import with.
        reset_modules (bool)[True]: If True reset sys.modules back to the original sys.modules.

    Returns:
        module (ModuleType): Module object that was imported.

    Raises:
        ImportError: If the import is unsuccessful.
    """
    if import_chain is None:
        import_chain, _ = get_name_version(path)
    if os.path.isfile(path):
        path = os.path.dirname(path)

    if os.path.exists(path):
        # Import the module
        with original_system(os.path.abspath(path), reset_modules=reset_modules):
            module = importlib.import_module(import_chain)  # module = __import__(name)
        return module


INSTALL_TYPES = OrderedDict()


def register_install_type(type_func, install_func=None):
    """Register an install function.

    Args:
        type_func (str/callable/function): Extension string '.py' or callable function that takes in
            a path, **kwargs and returns a bool if the path is installable by this install function.
        install_func (callable/function)[None]: If None decorator. Function that takes in a path, dest and installs
            the path to the destination.

    Returns:
         install_func (callable/function): Decorator or install_func
    """
    if install_func is None:
        def decorator(install_func):
            return register_install_type(type_func, install_func)
        return decorator

    INSTALL_TYPES[type_func] = install_func
    return install_func


def remove_install_type(type_func):
    """Remove the registered type_func for installing."""
    try:
        INSTALL_TYPES.pop(type_func)
    except:
        pass
    try:
        INSTALL_TYPES.remove(type_func)
    except:
        pass


def get_install_func(path, **kwargs):
    """Find the installation function for the given path and keyword arguments.

    Args:
        path (str): Path to the installable file.
        **kwargs (dict): Dictionary of keyword arguments passed into the registered type_func to see if it can install
            the path.

    Returns:
        install_func (callable/function)[None]: Function to install the path with.
            Should take in (path, dest, **kwargs) and return True if installed or False if directory existed.
    """
    ext = os.path.splitext(path)[-1]
    install_func = INSTALL_TYPES.get(ext, None)
    if install_func is None:
        install_func = INSTALL_TYPES.get(ext.lower(), None)
        if install_func is None:
            for type_func, install_func in INSTALL_TYPES.items():
                try:
                    if callable(type_func) and type_func(path, **kwargs):
                        return install_func
                except (ValueError, TypeError, AttributeError, Exception):
                    pass

            return None  # No install func found!

    # Return the install func
    return install_func


def install_lib(path, dest, **kwargs):
    """Try to install the given path to the destination.

    Args:
        path (str): Path to the file or folder to install.
        dest (str): Destination path.

    Returns:
        installed (bool): If True it was installed this time. If False directory already existed.
    """
    install_func = get_install_func(path, **kwargs)
    if callable(install_func):
        return install_func(path, dest, **kwargs)

    raise InstallError('Invalid install type! Cannot install the given path "{}"!'.format(path))


def is_python_package(directory, **kwargs):
    """Return if the given directory has an __init__.py."""
    return os.path.exists(os.path.join(directory, '__init__.py'))


@register_install_type(is_python_package)
@register_install_type('.pyc')
@register_install_type('.pyd')
@register_install_type('.py')
def py_install(path, dest, *args, **kwargs):
    """Return the normal python import.

    Args:
        path (str): Path to the file or folder to install.
        dest (str): Destination path.

    Returns:
        installed (bool): If True it was installed this time. If False directory already existed.
    """
    # Make the path exist in the target dir
    try:
        os.makedirs(dest, exist_ok=True)
    except:
        pass

    try:
        try:
            # Create symlink
            os.symlink(path, dest, target_is_directory=os.path.isdir(path))
        except OSError:
            if os.path.isdir(path):
                shutil.copytree(path, dest)
            else:
                shutil.copy(path, dest)
    except (ValueError, TypeError, AttributeError, OSError, Exception) as err:
        raise InstallError('Failed to install "{}"'.format(path)) from err
    return True


def is_zip(path, **kwargs):
    return tarfile.is_tarfile(path)


@register_install_type(is_zip)
@register_install_type('.tar.gz')
@register_install_type('.zip')
def zip_install(path, dest, *args, **kwargs):
    """Install .zip or .tar.gz files.

    Args:
        path (str): Path to the file or folder to install.
        dest (str): Destination path.

    Returns:
        installed (bool): If True it was installed this time. If False directory already existed.

    Raises:
        InstallError: If any part of the installation fails.
    """
    # Extract to import location.
    try:
        os.makedirs(dest, exist_ok=True)
    except:
        pass

    # Extract to zip
    try:
        shutil.unpack_archive(path, dest)
        # DO NOT NEED TO MOVE UP ONE DIRECTORY ANYMORE
        # if not any(p == name for p in os.listdir(dest)):
        #     # Move items up one directory
        #     for p in os.listdir(dest):
        #         nested_path = os.path.join(dest, p)
        #         for np in os.listdir(nested_path):
        #             shutil.move(os.path.join(nested_path, np), os.path.join(dest, np))
        #         # shutil.rmtree(nested_path)
    except (ValueError, TypeError, AttributeError, OSError, Exception) as err:
        raise InstallError('Failed to install "{}"'.format(path)) from err
    return True


@register_install_type('.whl')
def whl_install(path, dest, *args, pip=None, extra_install_args=None, install_dependencies=False, wait_func=None,
                reset_modules=True, **kwargs):
    """Import whl or zip files and return the installed module.

    Args:
        dest (str): Destination path.
        path (str): Path to the zip file.
        pip (callable/function): pip function to install with (Takes in parameters passed into pip). Should be able
            to take in any keyword arguments and ignore them.
        extra_install_args (list/str): List of extra parameters to pass into the pip install command.
            Note: the '--target' argument is already being used.
        install_dependencies (bool)[False]: If True .whl files will install dependencies into the install_dir.
        wait_func (callable/function)[None]: Function called while waiting for pip to finish (passed into pip).
        reset_modules (bool)[True]: If True reset sys.modules back to the original sys.modules.

    Returns:
        installed (bool): If True it was installed this time. If False directory already existed.
    """
    if pip is None:
        pip = pip_main
    if not extra_install_args:
        extra_install_args = []
    elif isinstance(extra_install_args, str):
        extra_install_args = [extra_install_args]
    if wait_func is None:
        wait_func = default_wait_func

    # Install the wheel file to the target directory
    try:
        os.makedirs(dest, exist_ok=True)
    except:
        pass
    with original_system(dest, reset_modules=reset_modules):
        args = ['install', '--target', dest] + extra_install_args + [path]
        if not install_dependencies:
            args.insert(1, '--no-deps')

        exitcode = pip(*args, wait_func=wait_func)
        if exitcode != 0:
            try:
                shutil.rmtree(dest)
            except (OSError, Exception):
                pass
            raise InstallError('Could not install using pip with arguments {}'.format(args))
        return True
    return False


# ===== Make the module callable =====
# https://stackoverflow.com/a/48100440/1965288  # https://stackoverflow.com/questions/1060796/callable-modules
MY_MODULE = sys.modules[__name__]


class InstallModule(MY_MODULE.__class__):

    def __call__(self, path, dest, **kwargs):
        """Try to install the given path to the destination.

        Args:
            path (str): Path to the file or folder to install.
            dest (str): Destination path.

        Returns:
            installed (bool): If True it was installed this time. If False directory already existed.
        """
        return install_lib(path, dest, **kwargs)

# Override the module make it callable
try:
    MY_MODULE.__class__ = InstallModule  # Override __class__ (Python 3.6+)
    MY_MODULE.__doc__ = InstallModule.__call__.__doc__
except (TypeError, Exception):
    # < Python 3.6 Create the module and make the attributes accessible
    sys.modules[__name__] = MY_MODULE = InstallModule(__name__)
    for ATTR in __all__:
        setattr(MY_MODULE, ATTR, vars()[ATTR])


if __name__ == '__main__':
    import argparse

    P = argparse.ArgumentParser(description='Install the path to the destination')

    P.add_argument('path', type=str, help="Path to the python module, zip, whl, or folder to install.")
    P.add_argument('dest', type=str, help="Destination to install to.")


    # .whl arguments
    P.add_argument('--extra_install_args', type=str, default=None, help='Extra pip installation arguments.')
    P.add_argument('--install_dependencies', default=True, action='store_true',
                   help='If given install the dependencies in the destination as well.')

    ARGS = P.parse_args()
    KWARGS = {'extra_install_args': ARGS.extra_install_args, 'install_dependencies': ARGS.install_dependencies}
    if install_lib(ARGS.path, ARGS.dest, **KWARGS):
        print('Installed successfully!')
    else:
        print('Path already existed and may have already been installed.')



import sys
try:
    from . import fix_pip_subprocess
except (ImportError, Exception):
    pass


from .utils import default_wait_func
from .binary import find_file, IterProcess, pip_bin
from .main_func import PIP_MAIN_FUNC, is_pip_main_available, pip_main, is_pip_proc_available, pip_proc, pip_proc_flag


__all__ = ['default_wait_func', 'find_file', 'IterProcess', 'pip_bin',
           'PIP_MAIN_FUNC', 'is_pip_main_available', 'pip_main', 'is_pip_proc_available', 'pip_proc', 'pip_proc_flag']


# ===== Make the module callable =====
# https://stackoverflow.com/a/48100440/1965288  # https://stackoverflow.com/questions/1060796/callable-modules
MY_MODULE = sys.modules[__name__]


class RunPipModule(MY_MODULE.__class__):

    RUN_PIP = staticmethod(pip_main)

    def __call__(self, *args, **kwargs):
        """Run a pip command using the pip main function.

        https://pip.pypa.io/en/stable/reference/

        Note:
            pip_main may spawn another process. This may make executables not work.

        Example:

            >>> pip_main('install', '-e', '../mylib')

        Args:
            *args (tuple/str): Arguments that are normally passed into pip.
            **kwargs (dict/object)[None]: Catch extra arguments

        Raises:
            EnvironmentError: If pip could not be imported

        Returns:
            exitcode (int): Process exit code
        """
        return self.RUN_PIP(*args, **kwargs)


# Override the module make it callable
try:
    MY_MODULE.__class__ = RunPipModule  # Override __class__ (Python 3.6+)
    MY_MODULE.__doc__ = RunPipModule.__call__.__doc__
except (TypeError, Exception):
    # < Python 3.6 Create the module and make the attributes accessible
    sys.modules[__name__] = MY_MODULE = RunPipModule(__name__)
    for ATTR in __all__:
        setattr(MY_MODULE, ATTR, vars()[ATTR])



from PyInstaller import config
from PyInstaller.utils.hooks import collect_data_files

# pip requirements if using python code (pip_proc or get_pip_main())
datas = collect_data_files('pip')
hiddenimports = [
    'setuptools',
    'distutils',
    'pkg_resources',
    ]

try:
    import pip._internal.commands
    hiddenimports.extend([v.module_path for v in pip._internal.commands.commands_dict.values()])
except (ImportError, AttributeError, Exception):
    hiddenimports.extend([
        'pip._internal.commands.install',
        'pip._internal.commands.download',
        'pip._internal.commands.uninstall',
        'pip._internal.commands.freeze',
        'pip._internal.commands.list',
        'pip._internal.commands.show',
        'pip._internal.commands.check',
        'pip._internal.commands.config',
        'pip._internal.commands.search',
        'pip._internal.commands.cache',
        'pip._internal.commands.wheel',
        'pip._internal.commands.hash',
        'pip._internal.commands.completion',
        'pip._internal.commands.debug',
        'pip._internal.commands.help',
        ])


# try:
#     from pylibimport import find_file
#
#     # Include the pip.exe file in your application
#     binaries = [(find_file('pip.exe', 'pip3', 'pip'), '.')]
#
# except (ImportError, Exception):
#     pass



import sys
from pylibimport.get_versions import HttpListVersions


__all__ = ['HttpListVersions']


# ===== Make the module callable =====
# https://stackoverflow.com/a/48100440/1965288  # https://stackoverflow.com/questions/1060796/callable-modules
MY_MODULE = sys.modules[__name__]


class DownloadModule(MY_MODULE.__class__):
    def __call__(self, package, version=None, download_dir='.', index_url='https://pypi.org/simple/', extensions=None,
                 min_version=None, exclude=None, check_compatibility=True, chunk_size=1024, **kwargs):
        """Return a series of package versions.

        Args:
            package (str): Name of the package/library you want to ge the versions for (Example: "requests").
            version (str)[None]: Version number to find and download.
            download_dir (str)['.']: Download directory.
            index_url (str) ['https://pypi.org/simple/']: Simple url to get the package and it's versions from.
            extensions (list/str) [None]: List of allowed extensions (Example: [".whl", ".tar.gz"]).
            min_version (str)[None]: Minimum version to allow.
            exclude (list)[None]: List of versions that are excluded.
            check_compatibility (bool)[True]: Check if the whl file works for this version of Python.
            chunk_size (int)[1024]: Save the file with this chunk size.

        Returns:
            filename (str): Filename of the downloaded file.
        """
        return HttpListVersions.download(package, version=version, download_dir=download_dir,
                                         index_url=index_url, extensions=extensions,
                                         min_version=min_version, exclude=exclude,
                                         check_compatibility=check_compatibility, chunk_size=chunk_size, **kwargs)

# Override the module make it callable
try:
    MY_MODULE.__class__ = DownloadModule  # Override __class__ (Python 3.6+)
    MY_MODULE.__doc__ = DownloadModule.__call__.__doc__
except (TypeError, Exception):
    # < Python 3.6 Create the module and make the attributes accessible
    sys.modules[__name__] = MY_MODULE = DownloadModule(__name__)
    for ATTR in __all__:
        setattr(MY_MODULE, ATTR, vars()[ATTR])


if __name__ == '__main__':
    import argparse

    P = argparse.ArgumentParser(description='List the versions for the given package')

    P.add_argument('package', type=str, help="Name of the package to list the versions for.")
    P.add_argument('-v', '--version', default=None, type=str, help="Version number you want to download.")
    P.add_argument('-o', '--download_dir', default='.', type=str, help="Directory to save the file to.")
    P.add_argument('-i', '--index_url', type=str, default='https://pypi.org/simple/', help='Index url to search.')
    P.add_argument('-e', '--extensions', metavar='N', type=str, nargs='+',
                   help='Allowed extensions (".whl", ".tar.gz")')
    P.add_argument('--min_version', type=str, default=None, help='Minimum version number to allow')
    P.add_argument('--check', '--check_compatibility', type=bool, default=True,
                   help='Check compatibility for this version of python.')
    P.add_argument('--exclude', metavar='N', type=str, nargs='*', help='Exclude versions')
    ARGS = P.parse_args()

    FILENAME = HttpListVersions.download(ARGS.package, version=ARGS.version, download_dir=ARGS.download_dir,
                                         index_url=ARGS.index_url, extensions=ARGS.extensions,
                                         min_version=ARGS.min_version, exclude=ARGS.exclude,
                                         check_compatibility=ARGS.check)
    print(FILENAME, 'saved!')



from .__meta__ import version as __version__

from .utils import make_import_name, get_name_version, get_compatibility_tags, \
    parse_filename, parse_wheel_filename, parse_sdist_filename, \
    is_compatible, get_meta, get_setup_dict

from .run_pip import default_wait_func, \
    find_file, IterProcess, pip_bin, \
    PIP_MAIN_FUNC, is_pip_main_available, pip_main, \
    is_pip_proc_available, pip_proc, pip_proc_flag

from .get_versions import uri_exists, HttpListVersions

from . import download

from .install import InstallError, original_system, import_module, install_lib, \
    register_install_type, remove_install_type, get_install_func, \
    is_python_package, py_install, is_zip, zip_install, whl_install

from .lib_import import VersionImporter


__all__ = [
    # meta
    '__version__'
    
    # utils
    'make_import_name', 'get_name_version', 'get_compatibility_tags',
    'parse_filename', 'parse_wheel_filename', 'parse_sdist_filename',
    'is_compatible', 'get_meta', 'get_setup_dict',

    # pip utils
    'default_wait_func',
    'find_file', 'IterProcess', 'pip_bin',
    'PIP_MAIN_FUNC', 'is_pip_main_available', 'pip_main',
    'is_pip_proc_available', 'pip_proc',

    # get versions
    'get_versions',   # Callable module
    'uri_exists', 'HttpListVersions',

    # download
    'download',  # Callable module

    # install
    'install',  # Callable module
    'InstallError', 'original_system', 'import_module', 'install_lib',
    'register_install_type', 'remove_install_type', 'get_install_func',
    'is_python_package', 'py_install', 'is_zip', 'zip_install', 'whl_install',

    # lib_import
    'lib_import',  # Callable module
    'VersionImporter',
    ]



import os
import sys
import requests
import contextlib
from collections import OrderedDict
from urllib.parse import urljoin
from urllib.request import urlopen
from html.parser import HTMLParser
from packaging.version import parse as parse_version

from pylibimport.utils import get_name_version, get_compatibility_tags, is_compatible


__all__ = ['uri_exists', 'HttpListVersions']


def uri_exists(uri):
    """Return if the given URI exists."""
    try:  # Faster check if url exists
        status_code = urlopen(uri).getcode()
        if 400 <= status_code < 500:
            raise ValueError('{} Client Error: Invalid url: {}'.format(status_code, uri))
        elif 500 <= status_code <= 600:
            raise ValueError('{} Server Error: Invalid url: {}'.format(status_code, uri))
        return True
    except (TypeError, ValueError, Exception):
        return False
    # try:
    #     with requests.get(str(uri), stream=True) as response:
    #         try:
    #             response.raise_for_status()
    #             return True
    #         except requests.exceptions.HTTPError:
    #             return False
    # except requests.exceptions.ConnectionError:
    #     return False


class HttpListVersions(HTMLParser):
    """Simple HTML parser to get the plugin, url names.

    Args:
        index_url (str) ['https://pypi.org/simple/']: Simple url to get the package and it's versions from.
        extensions (list/str) [None]: List of allowed extensions (Example: [".whl", ".tar.gz"]).
        min_version (str)[None]: Minimum version to allow.
        exclude (list)[None]: List of versions that are excluded.
    """

    EXTENSIONS = ['.whl', '.tar.gz', '.tar', '.zip', '.dist-info']
    is_compatible = staticmethod(is_compatible)

    def __init__(self, index_url='https://pypi.org/simple/', extensions=None, min_version=None, exclude=None,
                 check_compatibility=True, **kwargs):
        """Simple HTML parser to get the plugin, url names.

        Args:
            index_url (str) ['https://pypi.org/simple/']: Simple url to get the package and it's versions from.
            extensions (list/str) [None]: List of allowed extensions (Example: [".whl", ".tar.gz"]).
            min_version (str)[None]: Minimum version to allow.
            exclude (list)[None]: List of versions that are excluded.
            check_compatibility (bool)[True]: Check if the whl file works for this version of Python.
        """
        if extensions is None:
            extensions = self.EXTENSIONS
        elif not isinstance(extensions, (list, tuple)):
            extensions = [extensions]

        if index_url and not index_url.endswith('/'):
            index_url += '/'

        if exclude is None:
            exclude = []
        elif not isinstance(exclude, (list, tuple)):
            exclude = [exclude]
        exclude = list(exclude)

        self.index_url = index_url
        self.extensions = extensions
        self.check_compatibility = check_compatibility
        self._min_version = min_version
        self._min_version_parsed = parse_version(min_version or '-999.-999.-999')
        self._exclude = exclude
        self.saved_data = OrderedDict()
        super().__init__(**kwargs)

    @property
    def min_version(self):
        return self._min_version

    @min_version.setter
    def min_version(self, value):
        self._min_version = value
        if self._min_version is None:
            self._min_version_parsed = parse_version('-999.-999.-999')
        else:
            self._min_version_parsed = parse_version(self._min_version)

    @property
    def exclude(self):
        return self._exclude

    @exclude.setter
    def exclude(self, value):
        if value is None:
            value = []
        elif not isinstance(value, (list, tuple)):
            value = [value]
        value = list(value)
        self._exclude = value

    def uri_exists(self, index_url=None):
        """Return if the given URL/URI exists."""
        return uri_exists(index_url or self.index_url)

    def handle_starttag(self, tag, attrs):
        href = ''

        with contextlib.suppress(ValueError, TypeError, Exception):
            attrs = dict(attrs)
            href = attrs.get('href', '')

        if tag == 'a' and (len(self.extensions) == 0 or any(ext in href for ext in self.extensions)):
            if not self.check_compatibility or self.is_compatible(href):
                href = urljoin(self.index_url, href)
                name, version = get_name_version(href)
                if (version not in self.exclude and parse_version(version) >= self._min_version_parsed and
                        (name, version) not in self.saved_data):
                    self.saved_data[(name, version)] = href
                # print(name, version, href)

    def is_my_version(self, href):
        attrs = get_py_version(href)

    def handle_endtag(self, tag):
        pass

    def handle_data(self, data):
        pass

    @classmethod
    def href_as_filename(cls, href):
        """Return the href as a filename."""
        filename = href.rsplit('/')[-1]
        if "#" in filename:
            filename = filename.split('#')[0]
        return filename

    @classmethod
    def get_versions(cls, package, index_url='https://pypi.org/simple/', extensions=None,
                     min_version=None, exclude=None, check_compatibility=True, **kwargs):
        """Return a series of package versions.

        Args:
            package (str): Name of the package/library you want to ge the versions for (Example: "requests").
            index_url (str) ['https://pypi.org/simple/']: Simple url to get the package and it's versions from.
            extensions (list/str) [None]: List of allowed extensions (Example: [".whl", ".tar.gz"]).
            min_version (str)[None]: Minimum version to allow.
            exclude (list)[None]: List of versions that are excluded.
            check_compatibility (bool)[True]: Check if the whl file works for this version of Python.

        Returns:
            data (OrderedDict): Dictionary of {(package name, version): href}
        """
        parser = cls(index_url, extensions, min_version=min_version, exclude=exclude,
                     check_compatibility=check_compatibility, **kwargs)
        resp = requests.get(parser.index_url + package)
        if resp.status_code != 200:
            raise ValueError('Invalid URL.')

        parser.feed(resp.text)
        return parser.saved_data

    @classmethod
    def download(cls, package, version=None, download_dir='.', index_url='https://pypi.org/simple/', extensions=None,
                 min_version=None, exclude=None, check_compatibility=True, chunk_size=1024, **kwargs):
        """Return a series of package versions.

        Args:
            package (str): Name of the package/library you want to ge the versions for (Example: "requests").
            version (str)[None]: Version number to find and download.
            download_dir (str)['.']: Download directory.
            index_url (str) ['https://pypi.org/simple/']: Simple url to get the package and it's versions from.
            extensions (list/str) [None]: List of allowed extensions (Example: [".whl", ".tar.gz"]).
            min_version (str)[None]: Minimum version to allow.
            exclude (list)[None]: List of versions that are excluded.
            check_compatibility (bool)[True]: Check if the whl file works for this version of Python.
            chunk_size (int)[1024]: Save the file with this chunk size.

        Returns:
            filename (str): Filename of the downloaded file.
        """
        versions = cls.get_versions(package, index_url=index_url, extensions=extensions, min_version=min_version,
                                    exclude=exclude, check_compatibility=check_compatibility, **kwargs)
        if version is None:
            href = versions[list(versions)[-1]]  # Get latest version
        else:
            href = versions.get((package, version), None)
        if href is None:
            raise ValueError('Invalid version given. Version not found!')

        filename = cls.href_as_filename(href)
        filename = os.path.abspath(os.path.join(download_dir, filename))

        r = requests.get(href)

        with open(filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=chunk_size):
                f.write(chunk)

        return filename


# ===== Make the module callable =====
# https://stackoverflow.com/a/48100440/1965288  # https://stackoverflow.com/questions/1060796/callable-modules
MY_MODULE = sys.modules[__name__]


class GetVersionsModule(MY_MODULE.__class__):
    def __call__(self, package, index_url='https://pypi.org/simple/', extensions=None,
                 min_version=None, exclude=None, check_compatibility=True, **kwargs):
        """Return a series of package versions.

        Args:
            package (str): Name of the package/library you want to ge the versions for (Example: "requests").
            index_url (str) ['https://pypi.org/simple/']: Simple url to get the package and it's versions from.
            extensions (list/str) [None]: List of allowed extensions (Example: [".whl", ".tar.gz"]).
            min_version (str)[None]: Minimum version to allow.
            exclude (list)[None]: List of versions that are excluded.
            check_compatibility (bool)[True]: Check if the whl file works for this version of Python.

        Returns:
            data (OrderedDict): Dictionary of {(package name, version): href}
        """
        return HttpListVersions.get_versions(package, index_url=index_url, extensions=extensions,
                                             min_version=min_version, exclude=exclude,
                                             check_compatibility=check_compatibility, **kwargs)

# Override the module make it callable
try:
    MY_MODULE.__class__ = GetVersionsModule  # Override __class__ (Python 3.6+)
    MY_MODULE.__doc__ = GetVersionsModule.__call__.__doc__
except (TypeError, Exception):
    # < Python 3.6 Create the module and make the attributes accessible
    sys.modules[__name__] = MY_MODULE = GetVersionsModule(__name__)
    for ATTR in __all__:
        setattr(MY_MODULE, ATTR, vars()[ATTR])


if __name__ == '__main__':
    import argparse

    P = argparse.ArgumentParser(description='List the versions for the given package')

    P.add_argument('package', type=str, help="Name of the package to list the versions for.")
    P.add_argument('-i', '--index_url', type=str, default='https://pypi.org/simple/', help='Index url to search.')
    P.add_argument('-e', '--extensions', metavar='N', type=str, nargs='+',
                   help='Allowed extensions (".whl", ".tar.gz")')
    P.add_argument('--min_version', type=str, default=None, help='Minimum version number to allow')
    P.add_argument('--check', '--check_compatibility', type=bool, default=True,
                   help='Check compatibility for this version of python.')
    P.add_argument('--exclude', metavar='N', type=str, nargs='*', help='Exclude versions')
    ARGS = P.parse_args()

    VERSIONS = HttpListVersions.get_versions(ARGS.package, index_url=ARGS.index_url, extensions=ARGS.extensions,
                                             min_version=ARGS.min_version, exclude=ARGS.exclude,
                                             check_compatibility=ARGS.check)
    for (N, V), HREF in VERSIONS.items():
        print(N, V, HREF)



import os
import re
import sys

from packaging.tags import sys_tags, parse_tag
from packaging.utils import canonicalize_name, canonicalize_version
from packaging.version import Version, InvalidVersion


__all__ = ['make_import_name', 'get_name_version', 'get_compatibility_tags',
           'parse_filename', 'parse_wheel_filename', 'parse_sdist_filename',
           'is_compatible', 'get_meta', 'get_setup_dict']


def get_supported():
    tags = [(t.interpreter, t.abi, t.platform) for t in sys_tags()]
    for t in list(tags):
        if t[1].startswith('cp'):
            tags.append((t[0], t[1]+'m', t[2]))  # Copy tags with 'm' added to the abi
    return tags


SUPPORTED = get_supported()
_CONVERT_UNDERSCORE = re.compile("[^\w]+")


def make_import_name(name, version=''):
    """Return an import name using the name and version."""
    if version:
        converted_version = _CONVERT_UNDERSCORE.sub('_', str(version))
        if '.' in name:
            first, other = name.split('.', 1)
            return '{}_{}.{}'.format(first, converted_version, other)
        else:
            return '{}_{}'.format(name, converted_version)
    else:
        return name


def get_name_version(filename):
    """Parse a wheel filename.

    Args:
        filename (str): Wheel filename that ends in .whl or sdist filename that ends with .tar.gz.

    Returns:
        name (str): Name of the library
        version (str): Version of the library
    """
    if os.path.isdir(filename):
        filename = os.path.join(filename, 'setup.py')
    if filename.lower().endswith('setup.py'):
        filename = os.path.abspath(str(filename))
        try:
            meta = get_setup_dict(filename)
            return meta['name'], meta['version']
        except (ImportError, PermissionError, FileNotFoundError, KeyError, Exception):
            pass

        # Try finding the __meta__.py file to read. This is the only way I can read the proper name and version.
        filename = os.path.dirname(filename)
        for fname in os.listdir(filename):
            try:
                meta = get_meta(os.path.join(filename, fname, '__meta__.py'))
                return meta['name'], meta['version']
            except (FileNotFoundError, PermissionError, TypeError, KeyError, Exception):
                pass

    # Parse the wheel filename or sdist filename
    attrs = parse_filename(filename)
    return attrs['name'], str(attrs['version'])


def get_compatibility_tags(filename):
    """Get the python version and os architecture to check against.

    Args:
        filename (str): Wheel filename that ends in .whl or sdist filename that ends with .tar.gz.

    Returns:
        pyver (str)['py3']: Python version of the library py3 or cp38 ...
        abi (str)['none']: ABI version abi3, none, cp33m ...
        plat (str)['any']: Platform of the library none, win32, or win_amd64
    """
    # Parse the wheel filename or sdist filename
    attrs = parse_filename(filename)

    # Assume correct version if not found.
    return (attrs.get('pyver', 'py{}'.format(sys.version_info[0])),
            attrs.get('abi', 'none'),
            attrs.get('plat', 'any'))


def parse_filename(filename):
    """Parse a wheel filename or sdist filename and return the attributes.

    Args:
        filename (str): Wheel filename that ends in .whl or sdist filename that ends with .tar.gz.

    Returns:
        attrs (dict): Dictionary of attributes "name", "version", "build", "pyver", "abi", "plat".
    """
    attrs = {'name': '', 'version': '0.0.0', 'build': '', 'pyver': 'py{}'.format(sys.version_info[0]),
             'abi': 'none', 'plat': 'any'}

    # Format the filename
    filename = filename.split('#', 1)[0]  # Strip off md5 if applicable
    filename = os.path.basename(filename)

    try:
        # Use parse_wheel_filename
        values = parse_wheel_filename(filename)
        attrs['name'], attrs['version'], attrs['build'], attrs['pyver'], attrs['abi'], attrs['plat'] = values

    except (AttributeError, ValueError, Exception):
        try:
            # Use parse_sdist_filename
            values = parse_sdist_filename(filename)
            attrs['name'], attrs['version'] = values
        except (AttributeError, ValueError, Exception):
            fname = os.path.splitext(filename)[0]
            attrs['name'] = fname

            count = fname.count('-')
            if count >= 4:
                count = 4

            # Split the filename
            parts = fname.rsplit('-', count)

            # Find name and version
            if len(parts) >= 2:
                attrs['name'] = parts[0]
                attrs['version'] = parts[1]
            if '-' in attrs['name']:
                split = attrs['name'].split('-')
                attrs['name'] = split[0]
                attrs['version'] = split[1]

            # Find other
            py_ver_found = False
            for item in parts[2:]:  # Skip name and version
                if not py_ver_found and (item.startswith('py') or item.startswith('cp')):
                    py_ver_found = True
                    attrs['pyver'] = item
                elif (item.startswith('cp') or item.startswith('none') or item.startswith('abi')):
                    attrs['abi'] = item
                elif item.startswith('any') or item.startswith('win') or item.startswith('linux'):
                    attrs['plat'] = item

    return attrs


def parse_wheel_filename(filename):
    """Parse the wheel filename.

    Modified from: https://github.com/pypa/packaging/blob/master/packaging/utils.py
    """
    if not filename.endswith(".whl"):
        raise ValueError("Invalid wheel filename (extension must be '.whl'): {0}".format(filename))

    # Remove extension
    filename = filename[:-4]

    # Check number of dashes
    dashes = filename.count("-")
    if dashes < 4:  # Could always have more dashing in the "name" portion
        raise ValueError("Invalid wheel filename (wrong number of parts): {0}".format(filename))

    # Split the filename components
    name, version, pyver, abi, plat = filename.rsplit("-", 4)
    build = ''

    try:
        # Check if name contains versoin
        if '.' in name:
            raise ValueError  # name contains name and version while version is build?
        version = Version(version)
    except (InvalidVersion, ValueError, Exception):
        try:
            build = version
            name, version = name.rsplit('-', 1)
            version = Version(version)  # If this fails it is just invalid
        except (InvalidVersion, ValueError, Exception) as err:
            raise ValueError('Invalid version string "{}"!'.format(version)) from err

    # Check name formatting
    if "__" in name or re.match(r"^[\w\d._]*$", name, re.UNICODE) is None:
        raise ValueError("Invalid project name: {}".format(filename))

    name = canonicalize_name(name_part)
    return (name, version, build, pyver, abi, plat)


def parse_sdist_filename(filename):
    """Parse the sdist filename.

    Modified from: https://github.com/pypa/packaging/blob/master/packaging/utils.py
    """
    # Remove extension
    filename = os.path.splitext(filename)[0]  # '.tar.gz', '.zip', '.dist-info'
    if filename.endswith('.tar'):
        filename = os.path.splitext(filename)[0]

    # We are requiring a PEP 440 version, which cannot contain dashes,
    # so we split on the last dash.
    name, version = filename.rsplit('-', 1)
    name = canonicalize_name(name)
    version = Version(version)
    return (name, version)


def is_compatible(filename):
    """Return if the given filename is available on this system."""
    pyver, abi, plat = get_compatibility_tags(filename)
    return (pyver, abi, plat) in SUPPORTED


def get_meta(filename):
    """Return the metadata dictionary from the given filename."""
    with open(filename, 'r') as f:
        meta = {}
        exec(compile(f.read(), filename, 'exec'), meta)
        return meta


def get_setup_dict(filename):
    """Return the metadata dictionary from setup.py file."""
    meta = {}

    try:
        from setuptools import setup as orig_setup

        def my_setup(**attrs):
            meta.update(attrs)
        sys.modules['setuptools'].setup = my_setup

        cwd = os.getcwd()
        with open(filename, 'r') as f:
            os.chdir(os.path.abspath(os.path.dirname(filename)))
            exec(compile(f.read(), '<string>', 'exec'),
                 {'__name__': '__main__', '__file__': os.path.abspath(filename)})  # Hack __name__ and __file__ just in case
            os.chdir(cwd)

        sys.modules['setuptools'].setup = orig_setup

    except (ImportError, Exception):
        pass  # Failed to import

    return meta
